﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Windows.Forms;

namespace ArduinoHear
{
    public partial class Form1 : Form
    {

        #region Constants
        const int ss = 300; // скорость звука 300 м/с
        const double r = 0.1; // расстояние в метрах от точки O до точек A, B, C

        const double u30 = Math.PI / 6;
        const double u90 = Math.PI / 2;
        const double u120 = Math.PI * 2 / 3;
        const double u180 = Math.PI;
        const double u360 = Math.PI * 2;
        #endregion

        #region Init
        // центр экрана и масштаб отображения
        int cx;
        int cy;
        int scale;

        private Point pos = new Point(0, 0);
        private Point avg = new Point(0, 0);
        private List<Point> target = new List<Point>();
        private List<Sensor> sensors = new List<Sensor>();

        public Form1()
        {
            InitializeComponent();
            for (int i = 0; i < 3; i++)
            {
                Sensor s = new Sensor{position = new Point(r*Math.Cos(u120*i), r*Math.Sin(u120*i))};
                sensors.Add(s);
            }
            numericUpDown1_ValueChanged(null, null);
            panel1_Resize(null, null);
        }

        #endregion

        #region Privates

        private void DrawScene()
        {
            // инициация перерисовки сцены
            panel1.Refresh();
        }

        private void Draw(Graphics graphics)
        {
            // отрисовка сенсоров
            for (int i = 0; i < 3; i++)
                graphics.DrawLine(Pens.Blue, cx, cy, cx + (int) (scale*sensors[i].position.X), cy + (int) (scale*sensors[i].position.Y));
            // отрисовка вычисленных направлений
            foreach (Point point in target)
            {
                DrawPoint(graphics, Brushes.Green, point);
                graphics.DrawLine(Pens.Green, cx, cy, cx + (int)(scale * point.X), cy + (int)(scale * point.Y));
            }
            // отрисовка усредненного направления
            graphics.DrawLine(Pens.Red, cx, cy, cx + (int)(scale * avg.X), cy + (int)(scale * avg.Y));
            DrawPoint(graphics, Brushes.Red, avg);
            // отрисовка источника звука
            DrawPoint(graphics, Brushes.Black, pos);
        }

        private void DrawPoint(Graphics graphics, Brush brush, Point point)
        {
            graphics.FillEllipse(brush,
                cx - 2 + (int)(scale * point.X), cy - 2 + (int)(scale * point.Y), 
                5, 5);
        }

        private void Prepare()
        {
            double min = double.MaxValue;
            // вычисление времени распространения сигнала до сенсоров
            foreach (Sensor sensor in sensors)
            {
                double t = Distanse(sensor.position, pos)/ss;
                sensor.time = t;
                if (t < min) min = t;
            }
            // вычисление задержек между поступлением сигнала на сенсоры
            Random random = new Random(DateTime.Now.Millisecond);
            foreach (Sensor sensor in sensors)
            {
                sensor.time -= min;
                if (cbEnableErrors.Checked)
                    sensor.time = sensor.time * (1 + (double)random.Next(20) / 10000); // плюс погрешность не более 0.2%
            }
        }

        private void Calc()
        {
            // сортировка сенсоров по задержке получения сигнала (по удаленности)
            // sensors[0] = B, sensors[1] = A, sensors[2] = C
            int[] ix = new[] { 0, 1, 2 }; 
            if (sensors[ix[1]].time < sensors[ix[0]].time) swap(ix, 1, 0);
            if (sensors[ix[2]].time < sensors[ix[1]].time) swap(ix, 2, 1);
            if (sensors[ix[1]].time < sensors[ix[0]].time) swap(ix, 1, 0);
            // sensors[ix[0]] = B' - первым получил сигнал
            // sensors[ix[1]] = A' - вторым получил сигнал
            // sensors[ix[2]] = C' - третьим получил сигнал
            
            // определяем, как новый порядок перечисляет сенсоры: против часовой стрелки (1) или по часовой (-1)
            int orientation = (ix[0] + 1)%3 == ix[1] ? 1 : -1;

            // задание трех векторов(конец, начало), начало и конец вектора задается сенсором по его номеру
            int [][] v = new int[3][];
            v[0] = new[] { 0, 1 }; // вектор B'A'
            v[1] = new[] { 0, 2 }; // вектор B'C'
            v[2] = new[] { 2, 1 }; // вектор A'C'

            // углы отклонения трех векторов от O'B'
            double[] correction = new[] {-u30, u30, -u90};

            double avgAngle = 0;
            double prevAngle = 0;
            double prevLocalAngle = 0;
            target.Clear();

            // цикл по векторам
            for (int i = 0; i < 3; i++)
            {
                // вычисление угла между v[i] вектором и направлением на источник звука
                double localAngle = GetAngle(sensors[ix[v[i][0]]], sensors[ix[v[i][1]]]);
                // относительно вектора B'C' источник звука может оказаться как справа так и слева, поэтому ориентируемся на предыдущий угол
                if (i == 1 && prevLocalAngle < u30) localAngle = -localAngle;
                
                // вычисление угла между вектором OB и направлением на источник звука
                double angle = (localAngle + correction[i]) * orientation + u120 * ix[0];
                // корректировка угла относительно предыдущего с точностью до 2*PI
                Norm(prevAngle, ref angle);

                // вычисление среднего значения трех углов
                avgAngle += angle;

                // запоминаем значения как предыдущие
                prevLocalAngle = localAngle + correction[i];
                prevAngle = angle;

                // добавляем угол в список для отображения на графике 
                target.Add(new Point(Math.Cos(angle) * 5, Math.Sin(angle) * 5));
            }
            // вычисление среднего значения трех углов
            avgAngle /= 3;

            // вычисление расстояния

            double cos = Math.Cos(avgAngle);
            double sin = Math.Sin(avgAngle);
            Point A = sensors[ix[1]].position;
            double ta = ss*sensors[ix[1]].time;
            Point C = sensors[ix[2]].position;
            double tc = ss * sensors[ix[2]].time;
            // исследуемые границы 
            double low = r / 2; // от 5см 
            double hi = r * 1000; // до 1000м
            double eps = r / 10; // с точностью до 1 см.
            // подбор за 29 шагов, т.к. Log от 100000 по основанию 3/2 меньше 29
            while ((hi - low) > eps)
            {
                // вычисление точек внутри границы
                double l1 = (hi - low)/3;
                double l2 = hi - l1;
                l1 += low;
                // вычисление погрешности
                Point X = new Point(l1*cos, l1*sin);
                double err1 = GetError(X, C, tc, A, ta);
                X = new Point(l2 * cos, l2 * sin);
                double err2 = GetError(X, C, tc, A, ta);
                // сокращение границы на треть
                if (err1 > err2) low = l1;
                else hi = l2;
            }

            // получение координат точки через угол и расстояние от точки O
            avg = new Point(hi * Math.Cos(avgAngle), hi * Math.Sin(avgAngle));
        }

        private void Do(int x, int y)
        {
            // перевод экранных координат в пространственные
            pos.X = (double) (x - cx)/scale;
            pos.Y = (double) (y - cy)/scale;
            // подача сигнала на сенсоры
            Prepare();
            // вычисление направления
            Calc();
            // отрисовка сцены
            DrawScene();

            // вывод информации
            tbFactDist.Text = Distanse(pos, new Point()).ToString();
            tbCalcDist.Text = Distanse(avg, new Point()).ToString();
        }

        #endregion

        #region Static Math Methods

        private static double GetError(Point X, Point С, double tc, Point A, double ta)
        {
            double error = Math.Abs(Distanse(A, X) - ta - Distanse(С, X) + tc);
            return error;
        }

        private static double GetAngle(Sensor a, Sensor b)
        {
            double abDist = Distanse(a.position, b.position);
            double dtDist = (b.time - a.time)*ss;
            if (dtDist > abDist) dtDist = abDist;
            //if (dtDist < 0) dtDist = 0;
            double angle = Math.Acos(dtDist/abDist);
            return angle;
        }

        private static double Distanse(double x, double y)
        {
            return Math.Sqrt(x*x + y*y);
        }

        private static double Distanse(Point a, Point b)
        {
            return Distanse(a.X - b.X, a.Y - b.Y);
        }

        private static void swap(int[] mas, int a, int b)
        {
            int t = mas[a];
            mas[a] = mas[b];
            mas[b] = t;
        }

        private static void Norm(double prevAngle, ref double angle)
        {
            while (angle - prevAngle > u180)
                angle -= u360;
            while (angle - prevAngle < -u180)
                angle += u360;
        }

        #endregion

        #region Classes

        class Point
        {
            public double X, Y;
            public Point (double X, double Y)
            {
                this.X = X;
                this.Y = Y;
            }

            public Point()
            {
            }
        }

        class Sensor
        {
            public Point position = new Point();
            public double time;
        }

        #endregion

        #region Events

        private void panel1_Paint(object sender, PaintEventArgs e)
        {
            Draw(e.Graphics);
        }

        private void panel1_MouseDown(object sender, MouseEventArgs e)
        {
            if (rbByClick.Checked) Do(e.X, e.Y);
        }

        private void panel1_MouseMove(object sender, MouseEventArgs e)
        {
            if (rbByMove.Checked) Do(e.X, e.Y);
        }

        private void panel1_Resize(object sender, EventArgs e)
        {
            cx = panel1.Width/2;
            cy = panel1.Height/2;
            DrawScene();
        }

        private void numericUpDown1_ValueChanged(object sender, EventArgs e)
        {
            scale = (int)numericUpDown1.Value;
            DrawScene();
        }

        #endregion
    }
}
